# Add your Ansible Container service definitions here.
# For example:
  #
  # web:
  #   image: ubuntu:trusty
  #   ports:
  #     - "80:80"
  #   command: ['/usr/bin/dumb-init', '/usr/sbin/apache2ctl', '-D', 'FOREGROUND']
  #   dev_overrides:
  #     environment:
  #       - "DEBUG=1"
